@@ -50,6 +50,18 @@
 
 #include <linux/module.h>
 
+#ifdef CONFIG_HAVE_OLD_REQUEST_MODULE
+#include <linux/kmod.h>
+#undef request_module
+void snd_compat_request_module(const char *name, ...);
+#define request_module(name, args...) snd_compat_request_module(name, ##args)
+#endif
+
+#include <linux/compiler.h>
+#ifndef __user
+#define __user
+#endif
+
 #ifdef CONFIG_PCI
 #include <linux/pci.h>
 #endif
@@ -97,6 +109,12 @@
 #ifndef minor
 #define minor(x) MINOR(x)
 #endif
+#ifndef imajor
+#define imajor(x) major((x)->i_rdev)
+#endif
+#ifndef iminor
+#define iminor(x) minor((x)->i_rdev)
+#endif
 #ifndef mk_kdev
 #define mk_kdev(maj, min) MKDEV(maj, min)
 #endif
@@ -105,16 +123,27 @@
 	unsigned long name[((bits)+BITS_PER_LONG-1)/BITS_PER_LONG]
 #endif
 
-#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 3)
+#include <linux/sched.h>
+#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 3) && !defined(need_resched)
 #define need_resched() (current->need_resched)
 #endif
-#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 4)
+#ifndef CONFIG_HAVE_PDE
 #include <linux/fs.h>
+#include <linux/proc_fs.h>
 static inline struct proc_dir_entry *PDE(const struct inode *inode)
 {
 	return (struct proc_dir_entry *) inode->u.generic_ip;
 }
 #endif
+#ifndef cond_resched
+#define cond_resched() \
+	do { \
+		if (need_resched()) { \
+			set_current_state(TASK_RUNNING); \
+			schedule(); \
+		} \
+	} while (0)
+#endif
 #include <asm/io.h>
 #if !defined(isa_virt_to_bus)
 #if defined(virt_to_bus) || defined(__alpha__)
@@ -147,11 +176,19 @@ static inline struct proc_dir_entry *PDE(const struct inode *inode)
 #define MODULE_LICENSE(license)
 #endif
 
-/* no vsnprintf yet? */
-/* FIXME: the version number is not sure.. at least it exists already on 2.4.10 */
-#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 10)
-#define snprintf(buf,size,fmt,args...) sprintf(buf,fmt,##args)
-#define vsnprintf(buf,size,fmt,args) vsprintf(buf,fmt,args)
+#ifndef CONFIG_HAVE_STRLCPY
+size_t snd_compat_strlcpy(char *dest, const char *src, size_t size);
+#define strlcpy(dest, src, size) snd_compat_strlcpy(dest, src, size)
+size_t snd_compat_strlcat(char *dest, const char *src, size_t size);
+#define strlcat(dest, src, size) snd_compat_strlcat(dest, src, size)
+#endif
+
+#ifndef CONFIG_HAVE_SNPRINTF
+#include <stdarg.h>
+int snd_compat_snprintf(char * buf, size_t size, const char * fmt, ...);
+int snd_compat_vsnprintf(char *buf, size_t size, const char *fmt, va_list args);
+#define snprintf(buf,size,fmt,args...) snd_compat_snprintf(buf,size,fmt,##args)
+#define vsnprintf(buf,size,fmt,args) snd_compat_vsnprintf(buf,size,fmt,args)
 #endif
 
 #if defined(__alpha__) && LINUX_VERSION_CODE < KERNEL_VERSION(2, 3, 14)
@@ -166,19 +203,18 @@ static inline struct proc_dir_entry *PDE(const struct inode *inode)
 #define writeq(v, a) do { __writeq((v),(a)); mb(); } while(0)
 #endif
 
-#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 28)
 #include <linux/interrupt.h>
-#include <linux/sched.h>
+#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 28)
 static inline void synchronize_irq_wrapper(unsigned int irq) { synchronize_irq(); }
 #undef synchronize_irq
 #define synchronize_irq(irq)	synchronize_irq_wrapper(irq)
 #endif /* LINUX_VERSION_CODE < 2.5.28 */
-#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 68)
+#ifndef IRQ_NONE
 #define IRQ_NONE	/*void*/
 #define IRQ_HANDLED	/*void*/
 #define IRQ_RETVAL(x)	/*void*/
 typedef void irqreturn_t;
-#endif /* LINUX_VERSION_CODE < 2.5.68 */
+#endif
 
 #ifndef min
 /*
@@ -212,11 +248,14 @@ typedef void irqreturn_t;
 #undef devfs_remove
 void snd_compat_devfs_remove(const char *fmt, ...);
 #define devfs_remove snd_compat_devfs_remove
+#endif
 #if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 67)
 #undef devfs_mk_dir
-devfs_handle_t snd_compat_devfs_mk_dir(const char *dir);
+int snd_compat_devfs_mk_dir(const char *dir, ...);
 #define devfs_mk_dir snd_compat_devfs_mk_dir
-#endif
+#undef devfs_mk_cdev
+int snd_compat_devfs_mk_cdev(dev_t dev, umode_t mode, const char *fmt, ...);
+#define devfs_mk_cdev snd_compat_devfs_mk_cdev
 #endif
 #if LINUX_VERSION_CODE < KERNEL_VERSION(2, 3, 0)
 static inline void devfs_find_and_unregister (devfs_handle_t dir, const char *name,
@@ -242,7 +281,9 @@ static inline void devfs_find_and_unregister (devfs_handle_t dir, const char *na
 static inline void devfs_remove(const char *fmt, ...) { }
 #endif
 #undef devfs_mk_dir
-#define devfs_mk_dir(dir) do { (void)(dir); } while (0)
+#define devfs_mk_dir(dir, args...) do { (void)(dir); } while (0)
+#undef devfs_mk_cdev
+#define devfs_mk_cdev(dev, mode, fmt, args...) do { (void)(dev); } while (0)
 #endif /* CONFIG_DEVFS_FS */
 
 /* workarounds for USB API */
@@ -289,6 +330,7 @@ int snd_hack_usb_set_interface(struct usb_device *dev, int interface, int altern
 #define usb_host_config		usb_config_descriptor
 #define usb_host_interface	usb_interface_descriptor
 #define usb_host_endpoint	usb_endpoint_descriptor
+#define get_iface(cfg, num)	(&(cfg)->interface[num])
 #define get_iface_desc(iface)	(iface)
 #define get_endpoint(alt,ep)	(&(alt)->endpoint[ep])
 #define get_ep_desc(ep)		(ep)
@@ -351,11 +393,18 @@ struct page *snd_compat_vmalloc_to_page(void *addr);
 
 #if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 0) && LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 69)
 #include <linux/vmalloc.h>
+static inline void *snd_compat_vmap(struct page **pages, unsigned int count, unsigned long flags, pgprot_t prot)
+{
+	return vmap(pages, count);
+}
 #undef vmap
-void *snd_compat_vmap(struct page **pages, unsigned int count, unsigned long flags, pgprot_t prot);
 #define vmap snd_compat_vmap
 #endif
 
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 0) /* correct version? */
+#define EXPORT_NO_SYMBOLS
+#endif
+
 #include "amagic.h"
 
 #endif /* __SOUND_LOCAL_DRIVER_H */
